#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Collection of tools for numpy.array modifications. """

import textwrap
import warnings

import numpy as np

# =============================================================================
# Tools
# =============================================================================

def arrs_are_equal(arr1 ,arr2):
    """ True if arrays are the same. Also works for np.nan entries."""
    if arr1.shape == arr2.shape:
        return ((np.isnan(arr1) & np.isnan(arr2)) | (arr1 == arr2)).all()
    else:
        return False


def is_arr(data):
    if (isinstance(data,np.ndarray)):
        return True
    else:
        raise TypeError('Data needs to be np.ndarray.')


def check_time_dims(arr1, arr2):
    if (arr1.shape[0] == arr2.shape[0]):
        pass
    else:
        raise ValueError('Both input fields need to have same time dimensions.')


def remove_mean(arr):
    """Remove the mean of an array along the first dimension.

    If a variable (column) has at least 1 errorneous observation (row)
    the entire column will be set to NaN.

    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return arr - arr.mean(axis=0)


def remove_nan_cols(arr):
    """Remove NaN columns in array.

    Parameters
    ----------
    arr : ndarray
        Array to be cleaned.

    Returns
    -------
    new_data : ndarray
        Array without NaN columns.
    index : 1darray
        Index of columns without NaN entries from original data

    """
    " Remove all columns where the first row contains"
    index = np.where(~(np.isnan(arr[0])))[0]
    new_arr  = arr[:,index]

    return new_arr, index


def has_nan_time_steps(array):
    """ Checks if an array has NaN time steps.

    Time is assumed to be on axis=0. The array is then reshaped to be 2D with
    time along axis 0 and variables along axis 1. A NaN time step is a row which
    contain NaN only.
    """

    return (np.isnan(array).all(axis=tuple(range(1, array.ndim))).any())


def is_not_empty(arr):
    if (arr.size > 0):
        pass
    else:
        raise ValueError('Input field is empty or contains NaN only.')


def norm_to_1(arr, axis):
    return arr / np.nanmax(abs(arr), axis=axis)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Complex rotated maximum covariance analysis of two numpy arrays.
"""
# =============================================================================
# Imports
# =============================================================================
import cmath
import os
import warnings
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import hilbert
from statsmodels.tsa.forecasting.theta import ThetaModel
from tools.array import (arrs_are_equal, check_time_dims, has_nan_time_steps,
                         is_arr, is_not_empty, norm_to_1, remove_mean,
                         remove_nan_cols)
from tools.rotation import promax
from tools.text import boldify_str, secure_str, wrap_str
from tqdm import tqdm


# =============================================================================
# MCA
# =============================================================================
class MCA(object):
    """Perform Maximum Covariance Analysis (MCA) for two `numpy.ndarray`.

    MCA is a more general form of Principal Component Analysis (PCA)
    for two input fields (left, right). If both data fields are the same,
    it is equivalent to PCA.

    Parameters
    ----------
    left : ndarray
        Left input data. First dimension needs to be time.
    right : ndarray, optional
        Right input data. First dimension needs to be time.
        If none is provided, automatically, right field is assumed to be
        the same as left field. In this case, MCA reducdes to normal PCA.
        The default is None.


    Examples
    --------
    Let `data1` and `data2` be some geophysical fields (e.g. SST and pressure).
    To perform PCA use:

    >>> pca = CCA(data1)
    >>> pca.solve()
    >>> pcs,_ = pca.pcs()

    To perform MCA use:

    >>> mca = CCA(data1, data2)
    >>> mca.solve()
    >>> pcs_data1, pcs_data2 = mca.pcs()

    """

    def __init__(self, *data):
        """Load data fields and store information about data size/shape.

        Parameters
        ----------
        left : ndarray
            Left input data. First dimension needs to be time.
        right : ndarray, optional
            Right input data. First dimension needs to be time.
            If none is provided, automatically, right field is assumed to be
            the same as left field. In this case, MCA reducdes to normal PCA.
            The default is None.

        """
        if len(data) == 0:
            data = np.array([])

        if len(data) > 2:
            raise ValueError("Too many fields. Pass 1 or 2 fields.")

        if len(data) == 2:
            if data[0].shape[0] != data[1].shape[0]:
                raise ValueError("""Time dimensions of given fields are different.
                Time series should have same time lengths.""")

        if not all(isinstance(d, np.ndarray) for d in data):
            raise TypeError("""One or more fields are not `numpy.ndarray`.
            Please provide `numpy.ndarray` only.""")

        if any(has_nan_time_steps(d) for d in data):
            raise ValueError("""One or more fields contain NaN time steps.
            Please remove these prior to analysis.""")

        self._fields                = {} # input fields
        self._field_names           = {} # names of input fields
        self._field_means           = {} # mean of fields
        self._field_stds            = {} # standard deviation of fields
        self._fields_spatial_shape  = {} # spatial shapes of fields
        self._n_variables           = {} # number of variables of fields
        self._no_nan_index          = {} # index of variables containing data
        self._n_observations        = {} # number of observations/samples

        # set fields
        self._keys    = ['left', 'right']
        self._fields  = {key : field for key,field in zip(self._keys,data)}
        # set fields
        #self._fields = {'left'  : np.array([]) if left is None else left.copy()}
        #if right is not None:
        #    self._fields['right'] = right.copy()

        # set class variables
        for key, field in self._fields.items():
            self._field_names[key]          = key
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore',r'Mean of empty slice.')
                warnings.filterwarnings('ignore',r'invalid value encountered in double_scalars')
                warnings.filterwarnings('ignore',r'Degrees of freedom <= 0 for slice')
                warnings.filterwarnings('ignore',r'invalid value encountered in true_divide')
                self._field_means[key]          = field.mean(axis=0)
                self._field_stds[key]           = field.std(axis=0)
            self._fields_spatial_shape[key] = field.shape[1:]
            self._n_variables[key]          = np.product(field.shape[1:])
            self._n_observations[key]       = field.shape[0]

            # center input data to zero mean (remove mean)
            self._fields[key]       = remove_mean(field)

        # set meta information
        self._analysis = {
            'is_bivariate'          : True if len(self._fields) > 1 else False,
            # pre-processing
            'is_normalized'         : False,
            'is_coslat_corrected'   : False,
            'method'                : 'pca',
            # Complex solution
            'is_complex'            : False,
            'theta'                 : False,
            'theta_period'          : 365,
            #Rotated solution
            'is_rotated'            : False,
            'rotations'             : 0,
            'power'                 : 0,
            # Truncated solution
            'is_truncated'          : False,
            'is_truncated_at'       : 0,
            'rank'                  : 0,
            'total_covariance'      : 0.0,
            'total_squared_covariance'      : 0.0
            }

        self._analysis['method']        = self._get_method_id()


    def set_field_names(self, left = None, right = None):
        if left is not None:
            self._field_names['left']   = left
        if right is not None:
            self._field_names['right']   = right


    def _get_method_id(self):
        id = 'pca'
        if self._analysis['is_bivariate']:
            id = 'mca'
        return id


    def _get_complex_id(self):
        id = int(self._analysis['is_complex'])
        return 'c{:}'.format(id)


    def _get_rotation_id(self):
        id = self._analysis['rotations']
        return 'r{:02}'.format(id)


    def _get_power_id(self):
        id = self._analysis['power']
        return 'p{:02}'.format(id)


    def _get_analysis_id(self):
        method      = self._get_method_id()
        hilbert     = self._get_complex_id()
        rotation    = self._get_rotation_id()
        power       = self._get_power_id()

        analysis    = '_'.join([method,hilbert,rotation,power])
        return analysis


    def _get_analysis_path(self, path=None):
        base_path   = path
        if base_path is None:
            base_path = os.getcwd()

        base_folder = 'xmca'

        analysis_folder = '_'.join(self._field_names.values())
        analysis_folder = secure_str(analysis_folder)

        analysis_path   = os.path.join(base_path, base_folder, analysis_folder)

        if not os.path.exists(analysis_path):
            os.makedirs(analysis_path)

        return analysis_path


    def _get_fields(self, original_scale=False):
        std     = self._field_stds
        mean    = self._field_means
        fields  = self._fields

        new_fields = {}
        for key, field in fields.items():
            new_fields[key] = field
            if original_scale:
                if self._analysis['is_normalized']:
                    new_fields[key] *= std[key]
                new_fields[key] += mean[key]

        return new_fields


    def apply_weights(self,left=None, right=None):
        """Apply weights to data sets.

        Supplied weights are applied via broadcasting.

        Parameters
        ----------
        left : ndarray
            Weights for left data set.
        right : ndarray
            Weights for right data set.


        """
        field_items = self._fields.items()

        weights = {'left' : left, 'right' : right}
        weights.update({k : 1 if w is None else w for k,w in weights.items()})
        self._fields.update({k : field * weight for k,field in field_items})


    def normalize(self):
        """Normalize the input data to unit variance."""


        keys        = self._fields.keys()
        fields      = self._fields.values()
        fields_std  = self._field_stds.values()

        for key,field,std in zip(keys,fields,fields_std):
            self._fields[key] = field / std

        self._analysis['is_normalized'] = True
        self._analysis['is_coslat_corrected'] = False
        self._analysis['method'] = self._get_method_id()
        return None


    def _theta_forecast(self, series):
        period = self._analysis['theta_period']
        steps = len(series)

        # replace last value of series by a mean value
        # to avoid some extreme cases where the foecast starts at a single
        # which may happen for very noisy data
        # series[0]   = series[::period].mean()
        # series[-1]  = series[::-period].mean()

        model = ThetaModel(series, period=period, deseasonalize=True, use_test=False).fit()
        forecast = model.forecast(steps=steps, theta=20)

        return forecast


    def _extend_data(self, data):
        extendedData = [self._theta_forecast(col) for col in tqdm(data.T)]
        extendedData = np.array(extendedData).T

        return extendedData


    def _complexify_data(self, data):
        """Complexify data via Hilbert transform.

        Calculating Hilbert transform via scipy.signal.hilbert is done
        through Fast Fourier Transform. If the time series exhibits some
        non-periodic behaviour (e.g. a trend) the Hilbert transform
        produces extreme "legs" at the beginning/end of the time series.
        To encounter this issue, we can forecast/backcast the original time
        series via the Theta model before applying the Hilbert transform.
        Then, we only take the middle part of the Hilbert transform
        (corresponding to the original time series) which exhibits
        a dampened influence of the "legs".

        Parameters
        ----------
        data : ndarray
            Real input data which is to be transformed via Hilbert transform.
        theta : boolean, optional
            If True, input time series are extended via forecast/backcast to
            3 * original length. This helps avoiding boundary effects of FFT.
        period : int, optional
            Period used to extend time series via Theta model. Using daily
            data a period of 365 represents seasonal cycle. The default is 365.

        Returns
        -------
        ndarray
            Analytical signal of input data.

        """
        n_observations = self._n_observations['left']

        if self._analysis['theta']:
            forecast    = self._extend_data(data)
            backcast    = self._extend_data(data[::-1])[::-1]

            data = np.concatenate([backcast, data, forecast])

        # perform actual Hilbert transform of (extended) time series
        data = hilbert(data,axis=0)

        if self._analysis['theta']:
            # cut out the first and last third of Hilbert transform
            # which belong to the forecast/backcast
            data = data[n_observations:(2*n_observations)]
            data = remove_mean(data)

        return data


    def solve(self, complexify=False, theta=False, period=365, save_memory=False):
        """Solve eigenvalue equation by performing SVD on covariance matrix.

        Parameters
        ----------
        complexify : boolean, optional
            Use Hilbert transform to complexify the input data fields
            in order to perform complex PCA/MCA. Default is false.
        theta : boolean, optional
            If True, extend time series by fore/backcasting based on
            Theta model. New time series will have 3 * original length.
            Only used for complex time series (complexify=True).
            Default is False.
        period : int, optional
            Seasonal period used for Theta model. Default is 365, representing
            a yearly cycle for daily data.
        """
        if any([np.isnan(field).all() for field in self._fields.values()]):
            raise RuntimeError('Fields are empty. Did you forgot to load data?')


        self._analysis['is_complex']    = complexify
        self._analysis['theta']         = theta
        self._analysis['theta_period']  = period

        n_observations  = self._n_observations
        n_variables     = self._n_variables

        field_2d = {}
        for key,field in self._fields.items():
            # create 2D matrix in order to perform SVD
            field = field.reshape(n_observations[key], n_variables[key])

            # remove NaNs columns in data fields
            field, no_nan_index   = remove_nan_cols(field)

            # complexify input data via Hilbert transform
            if self._analysis['is_complex']:
                field = self._complexify_data(field)

            field_2d[key]           = field
            # save index of real variables for recontruction of original field
            self._no_nan_index[key] = no_nan_index


        # create covariance matrix
        if self._analysis['is_bivariate']:
            kernel = field_2d['left'].conjugate().T @ field_2d['right']
        else:
            kernel = field_2d['left'].conjugate().T @ field_2d['left']
        kernel = kernel / n_observations['left']

        self._V = {} # singular vectors (EOFs)
        self._L = {} # "loaded" singular vectors
        self._U = {} # projections // (PCs)

        # perform singular value decomposition
        try:
            if save_memory:
                import scipy as sp
                VLeft, singular_values, VTRight = sp.sparse.linalg.svd(kernel, full_matrices=False)
            VLeft, singular_values, VTRight = np.linalg.svd(kernel, full_matrices=False)
        except LinAlgError:
            raise LinAlgError("SVD failed. NaN entries may be the problem.")

        self._V['left'] = VLeft
        if self._analysis['is_bivariate']:
            self._V['right'] = VTRight.conjugate().T
        # free up some space
        del(VLeft)
        del(VTRight)

        S = np.sqrt(np.diag(singular_values) * n_observations['left'])
        Si = np.diag(1./np.diag(S))

        self._singular_values = singular_values

        for key, V in self._V.items():
            # "loaded" EOFs
            self._L[key] = V @ S
            # get PC scores by projecting fields on loaded EOFs
            self._U[key] = field_2d[key] @ V @ Si

        self._analysis['total_covariance'] = singular_values.sum()
        self._analysis['total_squared_covariance'] = (singular_values**2).sum()
        self._analysis['rank'] = singular_values.size
        self._analysis['is_truncated_at'] = singular_values.size


    def rotate(self, n_rot, power=1, tol=1e-5):
        """Perform Promax rotation on the first `n` EOFs.

        Promax rotation (Hendrickson & White 1964) is an oblique rotation which
        seeks to find `simple structures` in the EOFs. It transforms the EOFs
        via an orthogonal Varimax rotation (Kaiser 1958) followed by the Promax
        equation. If `power=1`, Promax reduces to Varimax rotation. In general,
        a Promax transformation breaks the orthogonality of EOFs and introduces
        some correlation between PCs.

        Parameters
        ----------
        n_rot : int
            Number of EOFs to rotate.
        power : int, optional
            Power of Promax rotation. The default is 1.
        tol : float, optional
            Tolerance of rotation process. The default is 1e-5.

        Raises
        ------
        ValueError
            If number of rotations are <2.

        Returns
        -------
        None.

        """
        if(n_rot < 2):
            print('`n_rot` must be >=2. Solution not rotated.')
            return None
        if(power<1):
            raise ValueError('`power` must be >=1')

        n_observations = self._n_observations['left']

        truncated_L = {}
        for key, L in self._L.items():
            truncated_L[key] = L[:,:n_rot]

        # rotate loadings (Cheng and Dunkerton 1995)
        combined_L = np.concatenate(list(truncated_L.values()))
        combined_L, R, Phi = promax(combined_L, power, maxIter=1000, tol=tol)

        rot_V   = {} # rotated singular vectors (EOFs)
        rot_L   = {} # rotated "loading"
        rot_U   = {} # rotated projections (PCs)
        weights = {} # weights associated to "loadings"

        rot_L['left']   = combined_L[:truncated_L['left'].shape[0],:]
        if self._analysis['is_bivariate']:
            rot_L['right']  = combined_L[truncated_L['left'].shape[0]:,:]

        for key, L in rot_L.items():
            # calculate variance/reconstruct "singular_values"
            w = np.linalg.norm(L, axis=0)
            # pull loadings from EOFs
            rot_V[key] 	=  L / w
            weights[key] = w

        if self._analysis['is_bivariate']:
            variance = weights['left'] * weights['right'] / n_observations
        else:
            variance = weights['left'] * weights['left'] / n_observations
        var_idx = np.argsort(variance)[::-1]

        # rotate PC scores
        # If rotation is orthogonal: R.T = R
        # If rotation is oblique (p>1): R^(-1).T = R
        for key, pcs in self._U.items():
            if(power==1):
                rot_U[key] = pcs[:,:n_rot] @ R
            else:
                rot_U[key] = pcs[:,:n_rot] @ np.linalg.pinv(R).conjugate().T

        # store rotated pcs, eofs and "singular_values"
        # and sort according to described variance
        self._singular_values 	= variance[var_idx]
        for key in rot_L.keys():
            self._V[key] = rot_V[key][:,var_idx] # Standardized EOFs
            self._L[key] = rot_L[key][:,var_idx] # Loaded EOFs
            self._U[key] = rot_U[key][:,var_idx] # Standardized PC scores

        # store rotation and correlation matrix of PCs + meta information
        self._rotation_matrix           = R
        self._correlation_matrix        = Phi[var_idx,:][:,var_idx]
        self._analysis['is_rotated']    = True
        self._analysis['rotations']     = n_rot
        self._analysis['power']         = power


    def rotation_matrix(self):
        """
        Return the rotation matrix.

        Returns
        -------
        ndarray
            Rotation matrix.
        """
        if (self._analysis['is_rotated']):
            return self._rotation_matrix
        else:
            print('Apply `.rotate()` first to retrieve the correlation matrix.')


    def correlation_matrix(self):
        """
        Return the correlation matrix of rotated PCs.

        Returns
        -------
        ndarray
            Correlation matrix.

        """
        if (self._analysis['is_rotated']):
            return self._correlation_matrix
        else:
            print('Apply `.rotate()` first to retrieve the correlation matrix.')


    def singular_values(self,n=None):
        """Return the first `n` singular_values.

        Parameters
        ----------
        n : int, optional
            Number of singular_values to return. The default is 5.

        Returns
        -------
        values : ndarray
            singular_values of PCA.
        error : ndarray
            Uncertainty of singular_values according to North's rule of thumb.

        """
        values = self._singular_values[:n]
        n_observations = self._n_observations['left']
        # error according to North's Rule of Thumb
        error = np.sqrt(2 / n_observations) * values

        return values

    def scf(self, n=None):
        """Return the SCF of the first `n` modes.

        The squared ovariance/correlation fraction (SCF) is a measure of
        importance of each mode. It is calculated as the
        squared singular values divided by the sum of squared singluar values.
        In contrast to CF, it is an invariant for CCA.

        Parameters
        ----------
        n : int, optional
            Number of modes to return. The default is None.

        Returns
        -------
        ndarray
            Fraction of described squared covariance/correlation of each mode.

        """
        values  = self.singular_values(n)
        scf = values**2 / self._analysis['total_squared_covariance'] * 100
        return scf


    def explained_variance(self, n=None):
        """Return the CF of the first `n` modes.

        The covariance/correlation fraction (CF) is a measure of
        importance of each mode. It is calculated as the
        singular values divided by the sum of singluar values.

        Parameters
        ----------
        n : int, optional
            Number of modes to return. The default is None.

        Returns
        -------
        ndarray
            Fraction of described covariance/correlation of each mode.

        """
        values  = self.singular_values(n)
        exp_var = values / self._analysis['total_covariance'] * 100
        return exp_var


    def pcs(self, n=None, scaling=None, phase_shift=0):
        """Return the first `n` PCs.

        Parameters
        ----------
        n : int, optional
            Number of PCs to be returned. The default is None.
        scaling : {None, 'eigen', 'max', 'std'}, optional
            Scale by singular_values ('eigen'), maximum value ('max') or
            standard deviation ('std'). The default is None.

        Returns
        -------
        pcs : dict[ndarray, ndarray]
            PCs associated to left and right input field.

        """
        n_obs       = self._n_observations['left']
        singular_values = self._singular_values

        if n is None:
            n = singular_values.size

        pcs = {}
        for key, U in self._U.items():
            pcs[key] = U[:,:n].copy()
            # apply phase shift
            if self._analysis['is_complex']:
                pcs[key] *= cmath.rect(1,phase_shift)
            # scale PCs by singular_values
            if scaling == 'eigen':
                pcs[key] *= np.sqrt(n_obs * singular_values[:n])
            # scale PCs by maximum value
            if scaling == 'max':
                pcs[key] /= np.nanmax(abs(pcs[key].real), axis=0)
            # scale PCs by standard deviation
            if scaling == 'std':
                pcs[key] /= np.nanstd(abs(pcs[key].real), axis=0)

        return pcs


    def eofs(self, n=None, scaling=None, phase_shift=0):
        """Return the first `n` EOFs.

        Parameters
        ----------
        n : int, optional
            Number of EOFs to be returned. The default is None.
        scaling : {None, 'eigen', 'max', 'std'}, optional
            Scale by singular_values ('eigen'), maximum value ('max') or
            standard deviation ('std'). The default is None.

        Returns
        -------
        eofs : dict[ndarray, ndarray]
            EOFs associated to left and right input field.

        """
        n_obs       = self._n_observations['left']
        n_var       = self._n_variables
        no_nan_idx  = self._no_nan_index
        field_shape = self._fields_spatial_shape
        singular_values = self._singular_values

        if n is None:
            n = self._singular_values.size

        eofs = {}
        for key, V in self._V.items():
            # create data fields with original NaNs
            dtype       = V.dtype
            eofs[key]   = np.zeros([n_var[key], n], dtype = dtype) * np.nan
            eofs[key][no_nan_idx[key],:] = V[:,:n]
            # reshape data fields to have original input shape
            eofs[key] 	= eofs[key].reshape(field_shape[key] + (n,))
            # scale EOFs with their singular_values
            if scaling == 'eigen':
                eofs[key] *= np.sqrt(n_obs * singular_values[:n])
            # scale EOFs by maximum value
            if scaling == 'max':
                eofs[key] /= np.nanmax(abs(eofs[key].real), axis=(0,1))
            # scale EOFs by standard deviation
            if scaling == 'std':
                eofs[key] /= np.nanstd(abs(eofs[key].real), axis=(0,1))
            # apply phase shift
            if self._analysis['is_complex']:
                eofs[key] *=  cmath.rect(1,phase_shift)

        return eofs


    def spatial_amplitude(self, n=None, scaling=None):
        """Return the spatial amplitude fields for the first `n` EOFs.

        Parameters
        ----------
        n : int, optional
            Number of amplitude fields to be returned. If None, return all
            fields. The default is None.
        scaling : {None, 'max', 'std'}, optional
            Scale by maximum value ('max'). The default is None.
        Returns
        -------
        amplitudes : dict[ndarray, ndarray]
            Spatial amplitude fields associated to left and right field.

        """
        eofs = self.eofs(n, scaling=None)

        amplitudes = {}
        for key, eof in eofs.items():
            amplitudes[key] = np.sqrt(eof * eof.conjugate()).real

            if scaling == 'max':
                amplitudes[key] /= np.nanmax(amplitudes[key], axis=(0,1))

        return amplitudes


    def spatial_phase(self, n=None, phase_shift=0):
        """Return the spatial phase fields for the first `n` EOFs.

        Parameters
        ----------
        n : int, optional
            Number of phase fields to return. If none, all fields are returned.
            The default is None.

        Returns
        -------
        amplitudes : dict[ndarray, ndarray]
            Spatial phase fields associated to left and right field.

        """
        eofs = self.eofs(n, phase_shift=phase_shift)

        phases = {}
        for key, eof in eofs.items():
            phases[key] = np.arctan2(eof.imag, eof.real).real

        return phases


    def temporal_amplitude(self, n=None, scaling=None):
        """Return the temporal amplitude time series for the first `n` PCs.

        Parameters
        ----------
        n : int, optional
            Number of amplitude series to be returned. If None, return all series.
            The default is None.
        scaling : {None, 'max'}, optional
            Scale by maximum value ('max'). The default is None.

        Returns
        -------
        amplitudes : dict[ndarray, ndarray]
            Temporal ampliude series associated to left and right field.

        """
        pcs = self.pcs(n, scaling=None)

        amplitudes = {}
        for key, pc in pcs.items():
            amplitudes[key] = np.sqrt(pc * pc.conjugate()).real

            if scaling == 'max':
                amplitudes[key] /= np.nanmax(amplitudes[key], axis=0)

        return amplitudes


    def temporal_phase(self, n=None, phase_shift=0):
        """Return the temporal phase function for the first `n` PCs.

        Parameters
        ----------
        n : int, optional
            Number of phase functions to return. If none, return all series.
            The default is None.

        Returns
        -------
        amplitudes : dict[ndarray, ndarray]
            Temporal phase function associated to left and right field.

        """
        pcs = self.pcs(n, phase_shift=phase_shift)

        phases = {}
        for key, pc in pcs.items():
            phases[key] = np.arctan2(pc.imag, pc.real).real

        return phases


    def plot(
        self, mode, threshold=0, phase_shift=0,
        cmap_eof=None, cmap_phase=None, figsize=(8.3,5.0)):
        """
        Plot results for `mode`.

        Parameters
        ----------
        mode : int, optional
            Mode to plot. The default is 1.
        threshold : int, optional
            Amplitude threshold below which the fields are masked out.
            The default is 0.
        cmap_eof : str or Colormap
            The colormap used to map the spatial patterns.
            The default is 'Blues'.
        cmap_phase : str or Colormap
            The colormap used to map the spatial phase function.
            The default is 'twilight'.

        Returns
        -------
        None.

        """
        pcs     = self.pcs(mode, scaling='max', phase_shift=phase_shift)
        eofs    = self.eofs(mode, scaling='max')
        phases  = self.spatial_phase(mode, phase_shift=phase_shift)
        var 	= self.explained_variance(mode)[mode-1]

        n_cols          = 2
        n_rows          = len(pcs)
        height_ratios   = [1] * n_rows

        # add additional row for colorbar
        n_rows += 1
        height_ratios.append(0.05)

        eof_title       = 'EOF'
        cmap_eof_range  = [-1, 0, 1]

        if self._analysis['is_complex']:
            n_cols          += 1
            eofs            = self.spatial_amplitude(mode, scaling='max')
            eof_title       = 'Amplitude'
            cmap_eof_range  = [0, 1]
            cmap_eof        = 'Blues' if cmap_eof is None else cmap_eof
            cmap_phase      = 'twilight' if cmap_phase is None else cmap_phase
        else:
            cmap_eof        = 'RdBu_r' if cmap_eof is None else cmap_eof

        for key in pcs.keys():
            pcs[key]    = pcs[key][:,mode-1].real
            eofs[key]   = eofs[key][:,:,mode-1]
            phases[key]  = phases[key][:,:,mode-1]

            # apply amplitude threshold
            eofs[key]  = np.where(abs(eofs[key]) >= threshold, eofs[key], np.nan)
            phases[key] = np.where(abs(eofs[key]) >= threshold, phases[key], np.nan)


        titles = {
        'pc' : 'PC {:d} ({:.1f} \%)'.format(mode,var),
        'eof': eof_title,
        'phase':'Phase',
        'var1' : self._field_names['left'],
        'var2' : self._field_names['right']
        }

        titles.update({k: v.replace('_',' ') for k, v in titles.items()})
        titles.update({k: boldify_str(v) for k, v in titles.items()})


        # create figure environment
        fig = plt.figure(figsize=figsize, dpi=150)
        fig.subplots_adjust(hspace=0.1, wspace=.1, left=0.25)
        gs = fig.add_gridspec(n_rows, n_cols, height_ratios=height_ratios)
        axes_pc = [fig.add_subplot(gs[i,0]) for i in range(n_rows-1)]
        axes_eof = [fig.add_subplot(gs[i,1]) for i in range(n_rows-1)]
        cbax_eof = fig.add_subplot(gs[-1,1])

        axes_space = axes_eof

        var_names = [titles['var1'], titles['var2']]

        # plot PCs
        for i, pc in enumerate(pcs.values()):
            axes_pc[i].plot(pc)
            axes_pc[i].set_ylim(-1.2,1.2)
            axes_pc[i].set_xlabel('')
            axes_pc[i].set_ylabel(var_names[i], fontweight='bold')
            axes_pc[i].set_title('')
            axes_pc[i].set_yticks([-1,0,1])
            axes_pc[i].spines['right'].set_visible(False)
            axes_pc[i].spines['top'].set_visible(False)

        axes_pc[0].xaxis.set_visible(False)
        axes_pc[0].set_title(titles['pc'], fontweight='bold')

        # plot EOFs
        for i, eof in enumerate(eofs.values()):
            cb_eof = axes_eof[i].imshow(
                eof, origin='lower',
                vmin=cmap_eof_range[0], vmax=cmap_eof_range[-1], cmap=cmap_eof)
            axes_eof[i].set_title('')

        plt.colorbar(cb_eof, cbax_eof, orientation='horizontal')
        cbax_eof.xaxis.set_ticks(cmap_eof_range)
        axes_eof[0].set_title(titles['eof'], fontweight='bold')

        # plot Phase function (if data is complex)
        if (self._analysis['is_complex']):
            axes_phase = [fig.add_subplot(gs[i,2]) for i in range(n_rows-1)]
            cbax_phase = fig.add_subplot(gs[-1,2])

            for i, phase in enumerate(phases.values()):
                cb_phase = axes_phase[i].imshow(
                    phase, origin='lower',
                    vmin=-np.pi, vmax=np.pi, cmap=cmap_phase)
                axes_phase[i].set_title('')

            plt.colorbar(cb_phase, cbax_phase, orientation='horizontal')
            cbax_phase.xaxis.set_ticks([-3.14,0,3.14])
            cbax_phase.set_xticklabels([r'-$\pi$','0',r'$\pi$'])

            for a in axes_phase:
                axes_space.append(a)

            axes_phase[0].set_title(titles['phase'], fontweight='bold')

        # add map features
        for a in axes_space:
            a.set_aspect('auto')
            a.xaxis.set_visible(False)
            a.yaxis.set_visible(False)

        # if more than 1 row, remove xaxis
        if (len(pcs) == 2):
            axes_pc[0].xaxis.set_visible(False)
            axes_pc[0].spines['bottom'].set_visible(False)


    def save_plot(self, mode, path=None, dpi=96, **kwargs):
        if path is None:
            path = self._get_analysis_path()

        mode_id = ''.join(['mode',str(mode)])
        format = '.png'
        file_name = '_'.join([self._get_analysis_id(),mode_id])
        file_path = os.path.join(path, file_name)
        fig, axes= self.plot(mode=mode, **kwargs)
        fig.subplots_adjust(left=0.06)
        plt.savefig(file_path + format, dpi=dpi)


    def truncate(self, n):
        """Truncate solution.

        Parameters
        ----------
        n : int
            Cut off after mode `n`.


        """
        if (n < self._singular_values.size):
            self._singular_values = self._singular_values[:n]

            for key in self._U.keys():
                self._U[key] = self._U[key][:,:n]
                self._V[key] = self._V[key][:,:n]

            self._analysis['is_truncated'] = True
            self._analysis['is_truncated_at'] = n


    def _create_info_file(self, path):
        sep_line = '\n#' + '-' * 79
        now  = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        file_header = (
            'This file contains information neccessary to load stored analysis '
            'data from xmca module.')

        path_output   = os.path.join(path, self._get_analysis_id())
        path_output = '.'.join([path_output,'info'])

        file = open(path_output,"w+")
        file.write(wrap_str(file_header))
        file.write('\n# To load this analysis use:')
        file.write('\n# from xmca.xarray import xMCA')
        file.write('\n# mca = xMCA()')
        file.write('\n# mca.load_analysis(PATH_TO_THIS_FILE)')
        file.write('\n')
        file.write(sep_line)
        file.write(sep_line)
        file.write('\n{:<20} : {:<57}'.format('created',now))
        file.write(sep_line)
        for key, name in self._field_names.items():
            file.write('\n{:<20} : {:<57}'.format(key, str(name)))
        file.write(sep_line)
        for key, info in self._analysis.items():
            if key in ['is_bivariate','is_complex', 'is_rotated','is_truncated']:
                file.write(sep_line)
            file.write('\n{:<20} : {:<57}'.format(key, str(info)))
        file.close()


    def _get_file_names(self, format):
        base_name = self._get_analysis_id()

        fields  = {}
        eofs    = {}
        pcs     = {}
        for key, variable in self._field_names.items():
            variable    = secure_str(variable)
            field_name  = '_'.join([base_name, variable])
            eof_name    = '_'.join([base_name, variable, 'eofs'])
            pc_name     = '_'.join([base_name, variable, 'pcs'])

            fields[key] = '.'.join([field_name, format])
            eofs[key]   = '.'.join([eof_name, format])
            pcs[key]    = '.'.join([pc_name, format])

        singular_values = '_'.join([base_name, 'singular_values'])
        singular_values = '.'.join([singular_values, format])

        file_names = {
            'fields'    : fields,
            'eofs'      : eofs,
            'pcs'       : pcs,
            'singular'  : singular_values
        }
        return file_names


    def _save_data(self, data_array, path, *args, **kwargs):
        raise NotImplementedError('only works for `xarray`')


    def _set_analysis(self, key, value):
        try:
            key_type = type(self._analysis[key])
        except KeyError:
            raise KeyError("Key `{}` not found in info file.".format(key))
        if key_type == bool:
            self._analysis[key] = (value == 'True')
        else:
            self._analysis[key] = key_type(value)


    def _set_info_from_file(self, path):

        info_file = open(path, 'r')
        lines = info_file.readlines()
        for line in lines:
            if (line[0] != '#'):
                key = line.split(':')[0]
                key = key.rstrip()
                if key in ['left', 'right']:
                    name = line.split(':')[1].strip()
                    self._field_names[key] = name
                if key in self._analysis.keys():
                    value = line.split(':')[1].strip()
                    self._set_analysis(key, value)
        info_file.close()


    def load_analysis(self, path, fields, eofs, pcs, singular_values):
        self._set_info_from_file(path)

        self._V                     = {}
        self._L                     = {}
        self._U                     = {}
        self._singular_values           = singular_values
        for key in eofs.keys():
            self._n_observations[key]       = pcs[key].shape[0]
            self._fields_spatial_shape[key] = eofs[key].shape[:-1]
            self._n_variables[key]          = np.product(eofs[key].shape[:-1])
            n_modes                         = eofs[key].shape[-1]
            S   = np.sqrt(np.diag(singular_values) * self._n_observations[key])
            Si  = np.diag(1./np.diag(S))

            eofs[key]    = eofs[key].reshape(self._n_variables[key], n_modes)
            VT, self._no_nan_index[key]   = remove_nan_cols(eofs[key].T)

            self._V[key]    = VT.T
            self._L[key] 	= self._V[key] @ S
            self._U[key]    = pcs[key]

            self._field_means[key]  = fields[key].mean(axis=0)
            self._field_stds[key]   = fields[key].std(axis=0)

            self._fields[key]       = fields[key] - fields[key].mean(axis=0)

        if self._analysis['is_normalized']:
            self.normalize()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Collection of tools for xarray.DataArray modifications. """

# =============================================================================
# Imports
# =============================================================================
import numpy as np
import xarray as xr

# =============================================================================
# Tools
# =============================================================================

def is_DataArray(data):
    """Check if data is of type `xr.DataArray`.

    Parameters
    ----------
    A : DataArray
        Input data.

    Returns
    -------
    bool
        Input data is of type `DataArray`.

    """
    if(isinstance(data,xr.DataArray)):
        pass
    else:
        raise TypeError("Data format has to be xarray.DatArray.")


def check_dims(da1, da2):
    if (da1.dims == da2.dims):
        pass
    else:
        raise ValueError("Dimensions of input data has to be the same.")


def get_attr(data_array, attr, fallback='undefined'):
    try:
        return data_array.attrs[attr]
    except KeyError:
        return fallback


def calc_temporal_corr(x, y):
    is_DataArray(x)
    is_DataArray(y)

    x = x - x.mean('time')
    y = y - y.mean('time')

    xy = (x*y).mean('time')
    sigx = x.std('time')
    sigy = y.std('time')

    return xy/sigx/sigy


def wrap_lon_to_180(da, lon='lon'):
    """
    Wrap longitude coordinates of DatArray to -180..179

    Parameters
    ----------
    da : DatArray
        object with longitude coordinates
    lon : string
        name of the longitude ('lon', 'longitude', ...)

    Returns
    -------
    wrapped : Dataset
        Another dataset array wrapped around.
    """

    # wrap 0..359 to -180..179
    da = da.assign_coords(lon=(((da[lon] + 180) % 360) - 180))

    # sort the data
    return da.sortby(lon)


def get_extent(data_array, central_longitude=0):
    try:
        data_array = wrap_lon_to_180(data_array)
        east 	= data_array.coords['lon'].min() + central_longitude + 0.001
        west 	= data_array.coords['lon'].max() + central_longitude - 0.001
        south 	= data_array.coords['lat'].min()
        north 	= data_array.coords['lat'].max()

        return [east, west, south, north]

    except KeyError:
        KeyError("Spatial coordinates need to be called `lon` and `lat`.")


def norm_space_to_1(data_array):
    try:
        return data_array / abs(data_array).max(['lon','lat'])
    except KeyError:
        KeyError("Spatial coordinates need to be called `lon` and `lat`.")


def norm_time_to_1(data_array):
    try:
        return data_array / abs(data_array).max(['time'])
    except KeyError:
        KeyError("Temporal coordinate needs to be called `time`.")


def split_complex(data_array):
    dataset = xr.Dataset(
        data_vars = {
            'real': data_array.real,
            'imag': data_array.imag},
        attrs = data_array.attrs)

    return dataset


def create_coords(*coords):
    dims = [c.name for c in coords]
    sizes = [c.size for c in coords]

    data_array = np.random.randint(1,size=sizes)
    data_array = xr.DataArray(
        data_array,
        dims = dims,
        coords = coords)

    return data_array.coords



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Implementation of VARIMAX and PROMAX rotation. """

# =============================================================================
# Imports
# =============================================================================
import numpy as np

# =============================================================================
# VARIMAX
# =============================================================================

def varimax(A, gamma=1, maxIter=1000, tol=1e-8):
    '''
    Perform (orthogonal) Varimax rotation with Kaiser normalization (Kaiser 1958).

    Parameters
    ----------
    A : array-like
        PCs to be rotated. Matrix A has shape n x p with (n: number of time steps; p: number of pcs)
    gamma : float, optional
        Parameter which determines the type of rotation performed: varimax (1), quartimax (0). Other values are possible. The default is 1.
    maxIter : TYPE, optional
        Number of iterations performed. The default is 500.
    tol : TYPE, optional
        Tolerance at which iteration process terminates. The default is 1e-10.

    Returns
    -------
    B : array-like
        Rotated matrix with same dimensions as A.
    R : array-like
        Rotation matrix.

    '''
    A = A.copy()
    n,p = A.shape
    # initialize rotation matrix
    R =    np.eye(p)
    d = 0

    # normalize the matrix
    # using sqrt of the sum of squares (Kaiser)
    h = np.sqrt(np.sum(A * A.conjugate(),axis=1))
    # A = np.diag(1./h) @ A
    A = (1./h)[:,np.newaxis] * A

    # seek for rotation matrix based on varimax criteria
    converged = False
    for i in range(maxIter):
        d_old = d
        basis = A @ R

        transformed = A.conjugate().T @ (basis**2 * basis.conjugate() - (gamma/n) *
                             (basis @ np.diag(np.sum(basis*basis.conjugate(),axis=0))))

        u,s,vh = np.linalg.svd(transformed)
        R = u @ vh
        d = np.sum(s)
        if abs(d-d_old)/d < tol:
            converged = True
            break

    if(not converged):
        print('Error: Rotation process did not converge!')
        A = np.empty(A.shape)
        A[:] = np.nan

    # de-normalize using broadcasting
    A = h[:,np.newaxis] * A

    # perform rotation
    B = A @ R
    return B, R

# =============================================================================
# PROMAX
# =============================================================================

def promax(A, power=1, maxIter=1000, tol=1e-8):
    '''
    Perform (oblique) Promax rotation with Kaiser normalizataion (Kaiser 1958).

    Parameters
    ----------
    A : array-like
        PCs to be rotated. Matrix A has shape n x p with (n: number of time steps; p: number of pcs)
    power : int, optional
        Power the Varimax rotated solution to be raised to. If 1, Promax equals Varimax. The default is 1.

    Returns
    -------
    B : array-like
        Rotated matrix with same dimensions as A.
    R : array-like
        Rotation matrix.
    phi : array-like
        Correlation matrix.

    '''
    X = A.copy()
    n, p = X.shape
    if p < 2:
        print('Cannot rotate 1 PC. No rotation performed.')
        return X, np.eye(n), X.conjugate().T @ X

    # first get varimax rotation
    X, R = varimax(X,maxIter=maxIter,tol=tol)

    # pre-normalization by communalities (sum of squared rows)
    h = np.sqrt(np.sum(X*X.conjugate(),axis=1))
    # use broadcasting
    X = (1./h)[:,np.newaxis] * X
    #X = np.diag(1./h) @ X

    # max-normalisation of columns
    Xn = X / np.max(abs(X), axis=0)

    # "Procustes" equation
    P = Xn * np.abs(Xn)**(power - 1)

    # fit linear regression model of "Procrustes" equation
    # see Richman 1986 for derivation
    L = np.linalg.inv(X.conjugate().T @ X) @ X.conjugate().T @ P

    # calculate diagonal of inverse square
    try:
        sigma_inv = np.diag(np.diag(np.linalg.inv(L.conjugate().T @ L)))
    except np.linalg.LinAlgError:
        sigma_inv = np.diag(np.diag(np.linalg.pinv(L.conjugate().T @ L)))

    # transform and calculate inner products
    L = L @ np.sqrt(sigma_inv)
    B = X @ L

    # post-normalization based on Kaiser
    B =  h[:,np.newaxis] * B

    R = R @ L

    # Correlation matrix
    L_inv = np.linalg.inv(L)
    phi = L_inv @ L_inv.conjugate().T

    return B, R, phi



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Complex rotated maximum covariance analysis of two xarray DataArrays.
"""

import cmath
# =============================================================================
# Imports
# =============================================================================
import os
from datetime import datetime

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.gridspec import GridSpec
from xmca.array import MCA
from tools.text import boldify_str, secure_str, wrap_str
from tools.xarray import (calc_temporal_corr, check_dims, create_coords,
                          get_attr, get_extent, is_DataArray, split_complex)

# =============================================================================
# xMCA
# =============================================================================

class xMCA(MCA):
    """Perform Maximum Covariance Analysis (MCA) for two `xarray.DataArray`.

    MCA is a more general form of Principal Component Analysis (PCA)
    for two input fields (left, right). If both data fields are the same,
    it is equivalent to PCA.

    Parameters
    ----------
    left : ndarray
        Left input data. First dimension needs to be time.
    right : ndarray, optional
        Right input data. First dimension needs to be time.
        If none is provided, automatically, right field is assumed to be
        the same as left field. In this case, MCA reducdes to normal PCA.
        The default is None.


    Examples
    --------
    Let `data1` and `data2` be some geophysical fields (e.g. SST and pressure).
    To perform PCA use:

    >>> pca = MCA(data1)
    >>> pca.solve()
    >>> pcs = pca.pcs()
    >>> pcs['left'].sel(mode=1).plot()

    To perform MCA use:

    >>> mca = MCA(data1, data2)
    >>> mca.solve()
    >>> pcsData1, pcsData2 = mca.pcs()
    """

    def __init__(self, *data):
        """Load data fields and store information about data size/shape.

        Parameters
        ----------
        left : ndarray
            Left input data. First dimension needs to be time.
            The default is None.
        right : ndarray, optional
            Right input data. First dimension needs to be time.
            If none is provided, automatically, right field is assumed to be
            the same as left field. In this case, MCA reducdes to normal PCA.
            The default is None.

        Returns
        -------
        None.

        """
        if len(data) > 2:
            raise ValueError("Too many fields. Pass 1 or 2 fields.")

        if not all(isinstance(d, xr.DataArray) for d in data):
            raise TypeError("""One or more fields are not `xarray.DataArray`.
            Please provide `xarray.DataArray` only.""")

        # set fields
        keys    = ['left', 'right']
        fields  = {keys[i] : field for i,field in enumerate(data)}

        # store meta information of DataArrays
        self._field_dims    = {} # dimensions of fields
        self._field_coords  = {} # coordinates of fields

        for key,field in fields.items():
            self._field_dims[key]   = field.dims
            self._field_coords[key] = field.coords

        # constructor of base class for numpy.ndarray
        if len(fields) == 0:
            fields = {'left': xr.DataArray([])}
        fields = {key : field.values for key, field in fields.items()}
        super().__init__(*fields.values())


    def _get_fields(self, original_scale=False):
        dims        = self._field_dims
        coords      = self._field_coords
        field_names = self._field_names
        fields      = super()._get_fields(original_scale=original_scale)

        for key in fields.keys():
            fields[key] = xr.DataArray(
                fields[key],
                dims = dims[key],
                coords = coords[key],
                name = field_names[key])

            if (original_scale & self._analysis['is_coslat_corrected']):
                fields[key] /= np.cos(np.deg2rad(coords[key]['lat']))

        return fields



    def apply_weights(self, **weights):
        fields = self._get_fields()

        try:
            for key, weight in weights.items():
                self._fields[key]  = (fields[key] * weight).data
        except KeyError:
            raise KeyError("Keys not found. Choose `left` or `right`")


    def apply_coslat(self):
        """Apply area correction to higher latitudes.

        """
        coords  = self._field_coords
        weights = {}
        for key, coord in coords.items():
            weights[key] = np.sqrt(np.cos(np.deg2rad(coord['lat'])))

        if (not self._analysis['is_coslat_corrected']):
            self.apply_weights(**weights)
            self._analysis['is_coslat_corrected'] = True
        else:
            print("Coslat correction already applied. Nothing was done.")


    def singular_values(self, n=None):
        """Return first `n` singular_values of the PCA.

        Parameters
        ----------
        n : int, optional
            Number of singular_values to return. If none, then all singular_values are returned.
            The default is None.

        Returns
        -------
        DataArray
            singular_values of PCA.
        DataArray
            Uncertainty of singular_values according to North's rule of thumb.

        """
        # for n=Nonr, all singular_values are returned
        values = super().singular_values(n)

        # if n is not provided, take all singular_values
        if n is None:
            n = values.size

        modes = list(range(1,n+1))
        attrs = {k: str(v) for k, v in self._analysis.items()}

        values = xr.DataArray(values,
            dims 	= ['mode'],
            coords 	= {'mode' : modes},
            name 	= 'singular values',
            attrs   = attrs)

        # error = xr.DataArray(error,
        #     dims 	= ['mode'],
        #     coords 	= {'mode' : modes},
        #     name 	= 'error eigenvalues',
        #     attrs   = attrs)

        return values


    def explained_variance(self, n=None):
        """Return the CF of the first `n` modes.

        The covariance fraction (CF) is a measure of
        importance of each mode. It is calculated as the singular
        values divided by the sum of singular values.

        Parameters
        ----------
        n : int, optional
            Number of modes to return. The default is None.

        Returns
        -------
        DataArray
            Fraction of described covariance of each mode.

        """
        variance 	= super().explained_variance(n)

        # if n is not provided, take all singular_values
        if n is None:
            n = variance.size

        modes = list(range(1,n+1))
        attrs = {k: str(v) for k, v in self._analysis.items()}

        variance = xr.DataArray(variance,
            dims 	= ['mode'],
            coords 	= {'mode' : modes},
            name 	= 'covariance fraction',
            attrs   = attrs)

        # error = xr.DataArray(error,
        #     dims 	= ['mode'],
        #     coords 	= {'mode' : modes},
        #     name 	= 'error explained variance',
        #     attrs   = attrs)

        return variance

    def scf(self, n=None):
        """Return the SCF of the first `n` modes.

        The squared covariance fraction (SCF) is a measure of
        importance of each mode. It is calculated as the squared singular
        values divided by the sum of squared singular values. In contrast to CF,
        SCF is invariant under CCA.

        Parameters
        ----------
        n : int, optional
            Number of modes to return. The default is None.

        Returns
        -------
        DataArray
            Fraction of described squared covariance of each mode.

        """
        variance 	= super().scf(n)

        # if n is not provided, take all singular_values
        if n is None:
            n = variance.size

        modes = list(range(1,n+1))
        attrs = {k: str(v) for k, v in self._analysis.items()}

        variance = xr.DataArray(variance,
            dims 	= ['mode'],
            coords 	= {'mode' : modes},
            name 	= 'squared covariance fraction',
            attrs   = attrs)

        # error = xr.DataArray(error,
        #     dims 	= ['mode'],
        #     coords 	= {'mode' : modes},
        #     name 	= 'error explained variance',
        #     attrs   = attrs)

        return variance


    def pcs(self, n=None, scaling=None, phase_shift=0):
        """Return first `n` PCs.

        Parameters
        ----------
        n : int, optional
            Number of PCs to return. If none, then all PCs are returned.
        The default is None.
        scaling : {None, 'eigen', 'max', 'std'}, optional
            Scale by singular_values ('eigen'), maximum value ('max') or
            standard deviation ('std'). The default is None.

        Returns
        -------
        DataArray
            PCs of left input field.
        DataArray
            PCs of right input field.

        """
        pcs = super().pcs(n, scaling, phase_shift)

        if n is None:
            n = self._singular_values.size

        modes = list(range(1,n+1))
        attrs = {k: str(v) for k, v in self._analysis.items()}

        coords      = self._field_coords
        field_names = self._field_names
        for key, pc in pcs.items():
            pcs[key] = xr.DataArray(
                data        = pc,
                dims        = ['time','mode'],
                coords      = {'time' : coords[key]['time'], 'mode' : modes},
                name        = ' '.join([field_names[key],'pcs']),
                attrs       = attrs)

        return pcs


    def eofs(self, n=None, scaling=None, phase_shift=0):
        """Return the first `n` EOFs.

        Parameters
        ----------
        n : int, optional
            Number of EOFs to return If none, all EOFs are returned.
            The default is None.
        scaling : {None, 'eigen', 'max', 'std'}, optional
            Scale by singular_values ('eigen'), maximum value ('max') or
            standard deviation ('std'). The default is None.

        Returns
        -------
        DataArray
            EOFs of left input field.
        DataArray
            EOFs of right input field.

        """
        eofs = super().eofs(n, scaling, phase_shift)

        if n is None:
            n = self._singular_values.size

        modes = list(range(1,n+1))
        attrs = {k: str(v) for k, v in self._analysis.items()}

        coords      = self._field_coords
        field_names = self._field_names
        for key, eof in eofs.items():
            eofs[key] = xr.DataArray(
                data    = eof,
                dims    = ['lat','lon','mode'],
                coords  = {
                    'lon' : coords[key]['lon'],
                    'lat' : coords[key]['lat'],
                    'mode' : modes},
                name    = ' '.join([field_names[key],'eofs']),
                attrs   = attrs
                )

        return eofs


    def spatial_amplitude(self, n=None, scaling=None):
        """Return the spatial amplitude fields for the first `n` EOFs.

        Parameters
        ----------
        n : int, optional
            Number of amplitude fields to return. If none, all fields are returned.
            The default is None.
        scaling : {None, 'max'}, optional
            Scale by maximum value ('max'). The default is None.

        Returns
        -------
        DataArray
            Fields of left input field.
        DataArray
            Fields of right input field.

        """
        amplitudes = super().spatial_amplitude(n, scaling)

        if n is None:
            n = self._singular_values.size

        modes = list(range(1,n+1))
        attrs = {k: str(v) for k, v in self._analysis.items()}
        coords      = self._field_coords
        field_names = self._field_names

        for key, amp in amplitudes.items():
            amplitudes[key] = xr.DataArray(
                data    = amp,
                dims    = ['lat','lon','mode'],
                coords  = {
                'lon' : coords[key]['lon'],
                'lat' : coords[key]['lat'],
                'mode' : modes},
                name    = ' '.join([field_names[key],'spatial amplitude']),
                attrs   = attrs
                )

        return amplitudes


    def spatial_phase(self, n=None, phase_shift=0):
        """Return the spatial phase fields for the first `n` EOFs.

        Parameters
        ----------
        n : int, optional
            Number of phase fields to return. If none, all fields are returned.
            The default is None.
        scaling : {None, 'max', 'std'}, optional
            Scale by maximum value ('max') or
            standard deviation ('std'). The default is None.

        Returns
        -------
        DataArray
            Fields of left input field.
        DataArray
            Fields of right input field.

        """
        eofs = self.eofs(n, phase_shift=phase_shift)

        attrs = {k: str(v) for k, v in self._analysis.items()}
        field_names = self._field_names
        phases = {}
        for key, eof in eofs.items():
            phases[key]         = np.arctan2(eof.imag,eof.real).real
            phases[key].name    = ' '.join([field_names[key],'spatial phase'])
            phases[key].attrs   = attrs

        return phases


    def temporal_amplitude(self, n=None, scaling=None):
        """Return the temporal amplitude functions for the first `n` PCs.

        Parameters
        ----------
        n : int, optional
            Number of amplitude functions to return. If none, all functions are returned.
            The default is None.
        scaling : {None, 'max'}, optional
            Scale by maximum value ('max'). The default is None.

        Returns
        -------
        DataArray
            Temporal amplitude function of left input field.
        DataArray
            Temporal amplitude function of right input field.

        """
        amplitudes = super().temporal_amplitude(n, scaling)

        if n is None:
            n = self._singular_values.size

        modes = list(range(1,n+1))
        attrs = {k: str(v) for k, v in self._analysis.items()}
        coords      = self._field_coords
        field_names = self._field_names

        for key, amp in amplitudes.items():
            amplitudes[key] = xr.DataArray(
                data    = amp,
                dims    = ['time','mode'],
                coords  = {'time' : coords[key]['time'], 'mode' : modes},
                name    = ' '.join([field_names[key],'temporal amplitude']),
                attrs   = attrs
                )

        return amplitudes


    def temporal_phase(self, n=None, phase_shift=0):
        """Return the temporal phase function for the first `n` PCs.

        Parameters
        ----------
        n : int, optional
            Number of phase functions to return. If none, all functions are returned.
            The default is None.

        Returns
        -------
        DataArray
            Temporal phase function of left input field.
        DataArray
            Temporal phase function of right input field.

        """
        pcs = self.pcs(n, phase_shift=phase_shift)

        attrs = {k: str(v) for k, v in self._analysis.items()}
        field_names = self._field_names

        phases = {}
        for key,pc in pcs.items():
            phases[key] = np.arctan2(pc.imag,pc.real).real
            name = ' '.join([field_names[key],'temporal phase'])
            phases[key].name  = name
            phases[key].attrs = attrs

        return phases


    def homogeneous_patterns(self, n=None, phase_shift=0):
        """
        Return left and right homogeneous correlation maps.

        Parameters
        ----------
        n : int, optional
            Number of patterns (modes) to be returned. If None then all patterns
            are returned. The default is None.

        Returns
        -------
        xr.DataArray
            Left homogeneous correlation maps.
        xr.DataArray
            Right homogeneous correlation maps.

        """

        fields  = self._get_fields()
        pcs     = self.pcs(n, phase_shift)

        field_names = self._field_names
        attrs = {k: str(v) for k, v in self._analysis.items()}
        hom_patterns = {}
        for key, field in fields.items():
            hom_patterns[key] = calc_temporal_corr(fields[key],pcs[key].real)
            name = ' '.join([field_names[key],'homogeneous patterns'])
            hom_patterns[key].name  = name
            hom_patterns[key].attrs = attrs

        return hom_patterns


    def heterogeneous_patterns(self, n=None, phase_shift=0):
        """
        Return left and right heterogeneous correlation maps.

        Parameters
        ----------
        n : int, optional
            Number of patterns (modes) to be returned. If None then all patterns
            are returned. The default is None.

        Returns
        -------
        xr.DataArray
            Left heterogeneous correlation maps.
        xr.DataArray
            Right heterogeneous correlation maps.

        """
        fields  = self._get_fields()
        pcs     = self.pcs(n, phase_shift)

        field_names = self._field_names
        attrs = {k: str(v) for k, v in self._analysis.items()}
        het_patterns = {}
        reverse = {'left' : 'right', 'right' : 'left'}
        for key, field in fields.items():
            try:
                het_patterns[key] = calc_temporal_corr(fields[key],pcs[reverse[key]].real)
            except KeyError:
                raise KeyError("Key not found. Two fields needed for heterogenous maps.")
            name = ' '.join([field_names[key],'heterogenous patterns'])
            het_patterns[key].name  = name
            het_patterns[key].attrs = attrs

        return het_patterns


    def reconstructed_fields(self, mode=slice(1,None)):
        eofs    = self.eofs(scaling=None)
        pcs     = self.pcs(scaling='eigen')
        coords  = self._field_coords
        std     = self._field_stds
        mean    = self._field_means

        rec_fields = {}
        for key in self._fields.keys():
            eofs[key]   = eofs[key].sel(mode=mode)
            pcs[key]    = pcs[key].sel(mode=mode)
            rec_fields[key] = xr.dot(pcs[key],eofs[key].conjugate(),dims=['mode'])
            rec_fields[key] = rec_fields[key].real

            if self._analysis['is_coslat_corrected']:
                rec_fields[key] /= np.sqrt(np.cos(np.deg2rad(coords[key]['lat'])))

            if self._analysis['is_normalized']:
                rec_fields[key] *= std[key]

            # add mean fields
            rec_fields[key]  += mean[key]

        return rec_fields


    def _create_gridspec(self, figsize=None, orientation='horizontal', projection=None):
        is_bivariate    = self._analysis['is_bivariate']
        is_complex      = self._analysis['is_complex']

        n_rows =  2 if is_bivariate else 1
        n_cols =  3 if is_complex else 2
        height_ratios   = [1] * n_rows
        width_ratios    = [1] * n_cols
        # add additional row for colorbar
        n_rows += 1
        height_ratios.append(0.05)

        # horiontal layout
        if orientation == 'horizontal':
            grid = {
                'pc'   : {'left': [0, 0]},
                'eof'  : {'left': [0, 1]}
                }

            # position for phase
            if is_complex:
                grid['phase'] = {'left':[0,2]}

            # positions for right field
            if is_bivariate:
                for k, panel in grid.items():
                    yx = panel.get('left')
                    grid[k]['right'] = [yx[0]+1,yx[1]]

            # positions for colorbars
            for k, panel in grid.items():
                if k in ['eof','phase']:
                    row_cb = len(panel)
                    col_cb = panel.get('left')[1]
                    grid[k]['cb'] = [row_cb,col_cb]

        # vertical layout
        if orientation == 'vertical':
            grid = {
                'pc'   : {'left': [-1, 1]},
                'eof'  : {'left': [0, 1]}
                }

            # position for phase
            if is_complex:
                grid['phase'] = {'left':[1,1]}

            # positions for right field
            if is_bivariate:
                for k, panel in grid.items():
                    yx = panel.get('left')
                    grid[k]['right'] = [yx[0], yx[1]+1]

            # positions for colorbars
            for k, panel in grid.items():
                if k in ['eof','phase']:
                    row, col = panel.get('left')
                    grid[k]['cb'] = [row,col-1]

            n_rows, n_cols = n_cols, n_rows
            height_ratios = n_rows * [1]
            width_ratios = n_cols * [1]
            width_ratios[0] = 0.05

        map_projs = {}
        for k1, data in grid.items():
            map_projs[k1] = {}
            for k2 in data.keys():
                map_projs[k1][k2] = None
                if k1 in ['eof', 'phase']:
                    if k2 in ['left', 'right']:
                        map_projs[k1][k2] = projection.get(k2, None)

        # create figure and axes
        fig = plt.figure(figsize=figsize, dpi=150)
        gs = fig.add_gridspec(
            n_rows, n_cols,
            height_ratios=height_ratios, width_ratios=width_ratios
            )
        axes = {}
        for key_data,data in grid.items():
            axes[key_data] = {}
            for key_pos,pos in data.items():
                row,col = pos
                axes[key_data][key_pos] = fig.add_subplot(gs[row,col], projection=map_projs[key_data][key_pos])


        return fig, axes


    def plot(
        self, mode, threshold=0, phase_shift=0,
        cmap_eof=None, cmap_phase=None, figsize=(8.3,5.0), resolution='110m',
        projection=None, c_lon=None, orientation='horizontal', land=True):
        """
        Plot results for `mode`.

        Parameters
        ----------
        mode : int, optional
            Mode to plot. The default is 1.
        threshold : [0,1], optional
            Threshold of max-normalised amplitude below which the fields are
            masked out. The default is 0.
        cmap_eof : str or Colormap
            The colormap used to map the spatial patterns.
            The default is `Blues`.
        cmap_phase : str or Colormap
            The colormap used to map the spatial phase function.
            The default is `twilight`.
        resolution : {None, '110m', '50m', '10m'}
            A named resolution to use from the Natural Earth dataset.
            Currently can be one of `110m`, `50m`, and `10m`. If None, no
            coastlines will be drawn. Default is `110m`.
        orientation : {'horizontal', 'vertical'}
            Orientation of the plot. Default is horizontal.

        Returns
        -------
        fig :
            Figure instance.
        axes :
            Dictionary of axes containing `pcs`, `eofs` and `phase`, if complex.

        """
        complex     = self._analysis['is_complex']
        bivariate   = self._analysis['is_bivariate']

        # Get data
        var 		= self.explained_variance(mode).sel(mode=mode).values
        pcs         = self.pcs(mode, scaling='max', phase_shift=phase_shift)
        eofs        = self.eofs(mode, scaling='max')
        phases      = self.spatial_phase(mode, phase_shift=phase_shift)
        if complex:
            eofs = self.spatial_amplitude(mode, scaling='max')

        # ticks
        ticks = {
            'pc'       : [-1, 0, 1],
            'eof'      : [0, 1] if complex else [-1,0,1],
            'phase'    : [-np.pi, 0, np.pi]
        }

        # tick labels
        tick_labels = {
            'phase' : [r'-$\pi$','0',r'$\pi$']
        }

        # colormaps
        cmaps = {
            'eof'      : 'Blues' if complex else 'RdBu_r',
            'phase'    : 'twilight'
        }

        if not cmap_eof is None:
            cmaps['eof'] = cmap_eof
        if not cmap_phase is None:
            cmaps['phase'] = cmap_phase

        # titles
        titles = {
            'pc'        : 'PC',
            'eof'       : 'Amplitude' if complex else 'EOF',
            'phase'     : 'Phase',
            'mode'      : 'Mode {:d} ({:.1f} \%)'.format(mode,var)
        }

        for key, name in self._field_names.items():
            titles[key] = name

        titles.update({k: v.replace('_',' ') for k, v in titles.items()})
        titles.update({k: boldify_str(v) for k, v in titles.items()})

        # map projections and boundaries
        map = {
            # center longitude of maps
            'c_lon'     : {'left' : c_lon, 'right' : c_lon},
            # projections pf maps
            'projection' : {
                'left':ccrs.PlateCarree(),
                'right':ccrs.PlateCarree()
                },
            # west, east, south, north limit of maps
            'boundaries' : {'left': None, 'right' : None}
        }
        if projection is not None:
            try:
                map['projection'].update(projection)
            except TypeError:
                map['projection'] = {k: projection for k in map['projection'].keys()}
        data_projection  = ccrs.PlateCarree()

        # pre-process data and maps
        for key in pcs.keys():
            pcs[key] = pcs[key].sel(mode=mode).real
            eofs[key] = eofs[key].sel(mode=mode)
            phases[key] = phases[key].sel(mode=mode)

            # apply amplitude threshold
            eofs[key]   = eofs[key].where(abs(eofs[key]) >= threshold)
            phases[key] = phases[key].where(abs(eofs[key]) >= threshold)

            # map boundaries as [east, west, south, north]
            c_lon = map['projection'][key].proj4_params['lon_0']
            map['boundaries'][key] = get_extent(eofs[key], c_lon)


        fig, axes = self._create_gridspec(figsize=figsize, orientation=orientation, projection=map['projection'])

        # plot PCs, EOFs, and Phase
        for i, key in enumerate(pcs.keys()):
            # plot PCs
            pcs[key].plot(ax=axes['pc'][key])
            axes['pc'][key].set_ylim(-1.2,1.2)
            axes['pc'][key].set_yticks([-1,0,1])
            axes['pc'][key].set_ylabel(titles[key], fontweight='bold')
            axes['pc'][key].set_xlabel('')
            axes['pc'][key].set_title('')
            axes['pc'][key].spines['right'].set_visible(False)
            axes['pc'][key].spines['top'].set_visible(False)

            # plot EOFs
            cb_eof = eofs[key].plot(
                ax=axes['eof'][key], transform=data_projection,
                vmin=ticks['eof'][0], vmax=ticks['eof'][-1], cmap=cmaps['eof'],
                add_colorbar = False)
            axes['eof'][key].set_extent(map['boundaries'][key], crs=data_projection)
            axes['eof'][key].set_title('')

            if resolution in ['110m','50m','10m']:
                axes['eof'][key].coastlines(lw = .4, resolution = resolution)
            if land:
                axes['eof'][key].add_feature(cfeature.LAND, color='#808080', zorder=0)
            axes['eof'][key].set_aspect('auto')

            plt.colorbar(cb_eof, axes['eof']['cb'], orientation=orientation)
            if orientation == 'horizontal':
                axes['eof']['cb'].xaxis.set_ticks(ticks['eof'])
            elif orientation == 'vertical':
                axes['eof']['cb'].yaxis.set_ticks(ticks['eof'])

            # plot Phase function
            if complex:
                cb_phase = phases[key].plot(
                    ax=axes['phase'][key], transform=data_projection,
                    vmin=ticks['phase'][0], vmax=ticks['phase'][-1],
                    cmap=cmaps['phase'], add_colorbar = False)
                axes['phase'][key].set_extent(map['boundaries'][key], crs=data_projection)
                axes['phase'][key].set_title('')

                plt.colorbar(cb_phase, axes['phase']['cb'], orientation=orientation)
                if orientation=='horizontal':
                    axes['phase']['cb'].xaxis.set_ticks(ticks['phase'])
                    axes['phase']['cb'].set_xticklabels(tick_labels['phase'])
                elif orientation == 'vertical':
                    axes['phase']['cb'].yaxis.set_ticks(ticks['phase'])
                    axes['phase']['cb'].set_yticklabels(tick_labels['phase'])

                if resolution in ['110m','50m','10m']:
                    axes['phase'][key].coastlines(lw = .4, resolution = resolution)
                if land:
                    axes['phase'][key].add_feature(cfeature.LAND, color='#808080', zorder=0)
                axes['phase'][key].set_aspect('auto')
                axes['phase']['left'].set_title(titles['phase'], fontweight='bold')


        # special tweaking of axes according to orientation
        if orientation == 'horizontal':
            # titles
            axes['pc']['left'].set_title(titles['pc'], fontweight='bold')
            axes['eof']['left'].set_title(titles['eof'], fontweight='bold')

            if bivariate:
                axes['pc']['left'].xaxis.set_visible(False)
                axes['pc']['left'].spines['bottom'].set_visible(False)

        elif orientation == 'vertical':
            # titles
            axes['pc']['left'].set_ylabel(titles['pc'], fontweight='bold')
            axes['pc']['left'].set_title('')
            axes['eof']['left'].set_title(titles['left'], fontweight='bold')
            axes['eof']['cb'].set_ylabel(titles['eof'], fontweight='bold')
            axes['eof']['cb'].yaxis.set_label_position('left')
            axes['eof']['cb'].yaxis.set_ticks_position('left')
            if bivariate:
                axes['pc']['right'].yaxis.set_visible(False)
                axes['pc']['right'].spines['left'].set_visible(False)
                axes['eof']['right'].set_title(titles['right'], fontweight='bold')

            if complex:
                axes['phase']['cb'].set_ylabel(titles['phase'], fontweight='bold')
                axes['phase']['left'].set_title('')
                axes['phase']['cb'].yaxis.set_label_position('left')
                axes['phase']['cb'].yaxis.set_ticks_position('left')



        fig.subplots_adjust(wspace=.1)
        fig.suptitle(titles['mode'], horizontalalignment='left')

        return fig, axes


    def _save_data(self, data_array, path, engine='h5netcdf', *args, **kwargs):
        analysis_path   = path
        analysis_name   = self._get_analysis_id()
        var_name        = secure_str('.'.join([data_array.name,'nc']))

        file_name   = '_'.join([analysis_name, var_name])
        output_path = os.path.join(analysis_path,file_name)

        invalid_netcdf = True
        if engine != 'h5netcdf':
            invalid_netcdf = False
        data_array.to_netcdf(
            path=output_path,
            engine=engine, invalid_netcdf=invalid_netcdf, *args, **kwargs
            )


    def save_analysis(self, path=None, engine='h5netcdf'):
        analysis_path = self._get_analysis_path(path)

        self._create_info_file(analysis_path)

        fields      = self._get_fields(original_scale = True)
        eofs        = self.eofs()
        pcs         = self.pcs()
        singular_values = self.singular_values()

        self._save_data(singular_values, analysis_path, engine)
        for key in pcs.keys():
            self._save_data(fields[key], analysis_path, engine)
            self._save_data(eofs[key], analysis_path, engine)
            self._save_data(pcs[key], analysis_path, engine)



    def load_analysis(self, path, engine='h5netcdf'):
        self._set_info_from_file(path)
        path_folder,_ = os.path.split(path)
        file_names = self._get_file_names(format='nc')

        path_eigen   = os.path.join(path_folder,file_names['singular'])
        singular_values = xr.open_dataarray(path_eigen, engine = engine).data

        fields  = {}
        pcs     = {}
        eofs    = {}
        for key in self._field_names.keys():
            path_fields   = os.path.join(path_folder,file_names['fields'][key])
            path_pcs   = os.path.join(path_folder,file_names['pcs'][key])
            path_eofs   = os.path.join(path_folder,file_names['eofs'][key])

            fields[key] = xr.open_dataarray(path_fields, engine = engine)
            pcs[key]    = xr.open_dataarray(path_pcs, engine = engine)
            eofs[key]   = xr.open_dataarray(path_eofs, engine = engine)


        self._field_coords = {}
        for key, field in fields.items():
            self._field_coords[key] = field.coords
            self._field_dims[key]   = field.dims

            fields[key]     = fields[key].data
            eofs[key]       = eofs[key].data
            pcs[key]        = pcs[key].data


        super().load_analysis(
            path = path,
            fields = fields,
            eofs = eofs,
            pcs  = pcs,
            singular_values = singular_values)

        if self._analysis['is_coslat_corrected']:
            self.apply_coslat()



from setuptools import setup, find_packages

"""
Run the following code in your conda environment to make the package available
$ python setup.py develop
"""


with open("README.md", "r") as fh:
    long_description = fh.read()


requirements = ['numpy >= 1.19.2',
                'xarray >= 0.16.2',
                'matplotlib >= 3.3.2',
                'statsmodels >= 0.12.2',
                'tqdm',
                'cartopy >= 0.18.0']


setup(
    name = "xmca",
    include_package_data = True,
    keywords = 'mca',
    author = "Niclas Rieger",
    author_email = "niclasrieger@gmail.com",
    description = "Maximum Covariance Analysis in Python",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/nicrie/xmca",
    packages = find_packages(),
    license = "GPL-3.0",
    classifiers = [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    version_config={
        "dev_template": "{tag}.post{ccount}",
        "dirty_template": "{tag}.post{ccount}"
    },
    setup_requires=['setuptools-git-versioning'],
    install_requires = requirements,
)
